/*
 * Copyright (c) Eugene Matyushkin 2007
 */
package ru.skipy.layout.circle;

import java.awt.*;
import static java.lang.Math.*;

/**
 * CircleLayout - layout manager implementation example. Lays out components on the circle in such way
 * that components will not intersect. Can fill whole container (layout components on ellipse rather then
 * circle), can use two directions of layout, can start from arbitrary angle specified.
 *
 * @author Eugene Matyushkin
 * @version $Id: CircleLayout.java 282 2007-10-22 07:49:54Z ematyushkin $
 */
public class CircleLayout implements LayoutManager2 {

    /**
     * Fill mode
     */
    private Fill fill;

    /**
     * Layout direction
     */
    private Direction direction;

    /**
     * Start angle from north direction clockwise (in radians)
     */
    private double startAngle = 0;


    /**
     * Constructs layout which does not fill container and directed clockwise starting from north.
     */
    public CircleLayout() {
        this(Fill.NONE, Direction.CLOCKWISE, 0);
    }


    /**
     * Constructs layout with the fill mode specified, which directed clockwise starting from north.
     *
     * @param fill fill mode
     */
    public CircleLayout(Fill fill) {
        this(fill, Direction.CLOCKWISE, 0);
    }


    /**
     * Constructs layout with the fill mode and direction specified, which performs layout starting from north.
     *
     * @param fill      fill mode
     * @param direction layout direction
     */
    public CircleLayout(Fill fill, Direction direction) {
        this(fill, direction, 0);
    }


    /**
     * Constructs layout.
     *
     * @param fill       fill mode
     * @param direction  layout direction
     * @param startAngle starting angle in radians from north direction clockwise
     */
    public CircleLayout(Fill fill, Direction direction, double startAngle) {
        if (fill == null) {
            throw new IllegalArgumentException("fill parameter can't be null!");
        }
        if (direction == null) {
            throw new IllegalArgumentException("direction parameter can't be null!");
        }
        this.fill = fill;
        this.direction = direction;
        this.startAngle = startAngle;
    }


    /**
     * Sets starting angle
     *
     * @param angle starting angle in radians from north direction clockwise
     */
    public void setStartAngle(double angle) {
        this.startAngle = angle;
    }


    /**
     * Sets fill mode
     *
     * @param fill fill mode
     */
    public void setFill(Fill fill) {
        this.fill = fill;
    }


    /**
     * Sets layout direction
     *
     * @param direction layout direction
     */
    public void setDirection(Direction direction) {
        this.direction = direction;
    }


    /**
     * Method from {@link java.awt.LayoutManager} interface. Lays out components within the container specified.
     *
     * @param parent container, that should be layed out
     */
    public void layoutContainer(Container parent) {
        if (parent.getComponentCount() == 0) {
            return;
        }
        if (parent.getComponentCount() == 1) {
            Component c = parent.getComponent(0);
            Dimension dp = c.getPreferredSize();
            Dimension ps = parent.getSize();
            c.setBounds((ps.width - dp.width) / 2, (ps.height - dp.height) / 2, dp.width, dp.height);
            return;
        }
        Dimension size = parent.getSize();
        double[] rd = getRadiusAndDeltaForContainer(parent);
        double rx = (fill == Fill.NONE || fill == Fill.VERTICAL) ? rd[0] : size.width / 2 - rd[1];
        if (rx < rd[0]) {
            rx = rd[0];
        }
        double ry = (fill == Fill.NONE || fill == Fill.HORIZONTAL) ? rd[0] : size.height / 2 - rd[1];
        if (ry < rd[0]) {
            ry = rd[0];
        }
        layoutComponents(parent.getComponents(), rx, ry, size);
    }


    /**
     * Lays out components using ellipse radiuses specified
     *
     * @param components components to layout
     * @param rx         ellipse radius along x axis
     * @param ry         ellipse radius along y axis
     * @param parentSize parent container size
     */
    private void layoutComponents(Component[] components, double rx, double ry, Dimension parentSize) {
        int cx = parentSize.width / 2;
        int cy = parentSize.height / 2;
        double alpha = direction.getCoeff() * 2 * PI / components.length;
        for (int i = 0; i < components.length; i++) {
            int ccx = cx + (int)(rx * sin(startAngle + i * alpha));
            int ccy = cy - (int)(ry * cos(startAngle + i * alpha));
            Dimension cs = components[i].getPreferredSize();
            components[i].setBounds(ccx - cs.width / 2, ccy - cs.height / 2, cs.width, cs.height);
        }
    }


    /**
     * Method from {@link java.awt.LayoutManager} interface. Calculates minimum layout size for the container
     * specified. Actually returns <code>new Dimension(0, 0)</code> when container is empty and calls
     * {@link #preferredLayoutSize(java.awt.Container)} otherwise.
     *
     * @param parent container, which minimum layout size should be calculated
     *
     * @return calculated minimum layout size
     */
    public Dimension minimumLayoutSize(Container parent) {
        if (parent.getComponentCount() == 0) {
            return new Dimension(0, 0);
        }
        return preferredLayoutSize(parent);
    }


    /**
     * Method from {@link java.awt.LayoutManager} interface. Calculates preferred layout size for the container
     * specified.
     *
     * @param parent container, which preferred layout size should be calculated
     *
     * @return calculated preferred layout size
     */
    public Dimension preferredLayoutSize(Container parent) {
        if (parent.getComponentCount() == 0) {
            return parent.getPreferredSize();
        }
        if (parent.getComponentCount() == 1) {
            return parent.getComponent(0).getPreferredSize();
        }
        double[] rd = getRadiusAndDeltaForContainer(parent);
        return new Dimension((int)(rd[0] + rd[1]) * 2 + 1, (int)(rd[0] + rd[1]) * 2 + 1);
    }


    /**
     * Calculates minimum layout radius and delta. Layout radius is a radius of the circle on which component
     * centers are placed. Delta is a difference between minimum enclosing circle's radius and layout radius.
     *
     * @param parent parent container
     *
     * @return array of 2 <code>float</code> elements - first if a radius, second is a delta.
     */
    private double[] getRadiusAndDeltaForContainer(Container parent) {
        Dimension[] dims = new Dimension[parent.getComponentCount()];
        Component[] components = parent.getComponents();
        for (int i = 0; i < components.length; i++) {
            dims[i] = components[i].getPreferredSize();
        }
        return calculateRadiusAndDelta(dims);
    }


    /**
     * Calculates minimum radius and delta for the component dimensions specified. Delta is a half of maximum
     * component's diagonal, radius is a maxumim of pair radiuses calculated by
     * {@link #calculateRadiusForPair(java.awt.Dimension,java.awt.Dimension,double,double)}
     *
     * @param components array of components' dimensions
     *
     * @return array of 2 <code>float</code> elements - first if a radius, second is a delta.
     */
    private double[] calculateRadiusAndDelta(Dimension[] components) {
        double[] res = new double[]{0, 0};
        if (components.length < 2) {
            return res;
        }
        for (int i = 0; i < components.length; i++) {
            double pr = calculateRadiusForPair(components[i], components[(i + 1 < components.length) ? i + 1 : 0], i,
                components.length);
            if (res[0] < pr) {
                res[0] = pr;
            }
            double mr = Math
                .sqrt(components[i].width * components[i].width + components[i].height * components[i].height) / 2.0;
            if (res[1] < mr) {
                res[1] = mr;
            }
        }
        return res;
    }


    /**
     * Calculates minimum radius of the circle for which components will not intersect if their centers will be placed
     * on this circle
     *
     * @param first      first component dimension
     * @param second     second component dimension
     * @param firstIndex first component index - required to calculate it's angle from start angle
     * @param total      total number of components - required to calculate angle between two subsequent components
     *
     * @return minimum radius
     */
    private double calculateRadiusForPair(Dimension first, Dimension second, double firstIndex, double total) {
        double alpha = direction.getCoeff() * 2. * PI / total;
        double beta = alpha * firstIndex + startAngle;
        double dSin = abs(sin(beta) - sin(alpha + beta));
        double dCos = abs(cos(beta) - cos(alpha + beta));
        double rx = (dSin < 1e-6) ? Double.POSITIVE_INFINITY : ((double)(first.width + second.width + 3)) / (2. * dSin);
        double ry =
            (dCos < 1e-6) ? Double.POSITIVE_INFINITY : ((double)(first.height + second.height + 3)) / (2. * dCos);
        return min(rx, ry);
    }


    /**
     * Method from {@link java.awt.LayoutManager2} interface. Adds component with the constraints specified
     * to the layout. Does nothing in this implementation.
     *
     * @param comp        component to add
     * @param constraints constraints to apply
     */
    public void addLayoutComponent(Component comp, Object constraints) {
    }


    /**
     * Method from {@link java.awt.LayoutManager} interface. Removes component from the layout.
     * Does nothing in this implementation.
     *
     * @param comp component to remove
     */
    public void removeLayoutComponent(Component comp) {
    }


    /**
     * Method from {@link java.awt.LayoutManager2} interface. Returns the alignment along the x axis for the
     * target container.
     *
     * @param target target container
     */
    public float getLayoutAlignmentX(Container target) {
        return 0.5f;
    }


    /**
     * Method from {@link java.awt.LayoutManager2} interface. Returns the alignment along the y axis for the
     * target container.
     *
     * @param target target container
     */
    public float getLayoutAlignmentY(Container target) {
        return 0.5f;
    }


    /**
     * Method from {@link java.awt.LayoutManager2} interface. Drops cached information if layout have it.
     */
    public void invalidateLayout(Container target) {
    }


    /**
     * Method from {@link java.awt.LayoutManager2} interface. Calculates maximumlayout size for the container
     * specified. Actually returns <code>new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE)</code>.
     *
     * @param target target container
     */
    public Dimension maximumLayoutSize(Container target) {
        return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
    }


    /**
     * Method from {@link java.awt.LayoutManager} interface. Adds component with the string association specified
     * to the layout. Some layouts may not support string associations.
     *
     * @param comp componnet to add
     * @param name string to associate with the component
     */
    public void addLayoutComponent(String name, Component comp) {
    }


    /**
     * Enumeration which defines filling mode fo the layout
     */
    public static enum Fill {

        /**
         * Layout does not fill the container
         */
        NONE,
        /**
         * Layout fills the container along horizontal axis
         */
        HORIZONTAL,
        /**
         * Layout fills the container along vertical axis
         */
        VERTICAL,

        /**
         * Layout fills the container along both axis
         */
        BOTH
    }

    /**
     * Enumeration which defines layout direction
     */
    public static enum Direction {

        /**
         * Clockwise direction
         */
        CLOCKWISE(1),
        /**
         * Counterclockwise direction
         */
        COUNTERCLOCWISE(-1);

        /**
         * direction multiplier
         */
        private int coeff;


        /**
         * Constructs enumeration entry with the direction multiplier specified
         *
         * @param coeff direction multiplier
         */
        private Direction(int coeff) {
            this.coeff = coeff;
        }


        /**
         * Returns direction multiplier
         *
         * @return direction multiplier
         */
        int getCoeff() {
            return coeff;
        }

    }

}
